###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Date, load_system_configuration
from .bbg_exceptions import BloombergFatal

from . import bbg_com_api as bbg_com_api

from collections import namedtuple

import logging
import zmq
import threading
import multiprocessing
import time
import traceback
import memcache
import hashlib

__all__ = ["BbgDataServer"]

NPROCS = 4  # number of worker processes
LOCK_TIMEOUT = 10  # in seconds
RT_TIMEOUT = 120  # in seconds

SORTED_KEYS = ["securities", "security", "fields",
               "start" "end", "adj", "overrides", "type", "RT"]

# --- define a convenience namedtuple that represents log messages
logmsg = namedtuple("logmsg", "level msg exc_info extra")


###############################################################################
class CacheLocked(object):
    __slots__ = ()


###############################################################################
class logmsg_pusher(object):
    # -------------------------------------------------------------------------
    def __init__(self, url):
        self.socket = zmq.Context().socket(zmq.PUSH)
        self.socket.connect(url)

    # -------------------------------------------------------------------------
    def info(self, msg, exc_info=None, extra=None):
        if exc_info is not None:
            # --- send the exception/traceback printout rather than the objects
            #     themselves
            exc_info = (traceback.format_exc(), )
        self.socket.send_pyobj(logmsg("info", msg, exc_info, extra))

    # -------------------------------------------------------------------------
    def debug(self, msg, exc_info=None, extra=None):
        if exc_info is not None:
            # --- send the exception/traceback printout rather than the objects
            #     themselves
            exc_info = (traceback.format_exc(), )
        self.socket.send_pyobj(logmsg("debug", msg, exc_info, extra))

    # -------------------------------------------------------------------------
    def warning(self, msg, exc_info=None, extra=None):
        if exc_info is not None:
            # --- send the exception/traceback printout rather than the objects
            #     themselves
            exc_info = (traceback.format_exc(), )
        self.socket.send_pyobj(logmsg("warning", msg, exc_info, extra))

    # -------------------------------------------------------------------------
    def critical(self, msg, exc_info=None, extra=None):
        if exc_info is not None:
            # --- send the exception/traceback printout rather than the objects
            #     themselves
            exc_info = (traceback.format_exc(), )
        self.socket.send_pyobj(logmsg("critical", msg, exc_info, extra))


# -----------------------------------------------------------------------------
def worker(wid, conn_str, mc_servers):
    # --- connect to master logger
    logger = logmsg_pusher("tcp://127.0.0.1:10001")
    logger.info("starting worker {0:d}".format(wid))

    # --- initialize memcached client
    cache = memcache.Client(mc_servers)

    # --- initialize bloomberg API for this thread
    bbg_com_api.BloombergApiInit()

    # --- connect to bloomberg data queue
    socket = zmq.Context().socket(zmq.REP)
    socket.connect(conn_str)

    while True:
        request = socket.recv_pyobj()

        logger.debug("{0:d}  request {1!r} "
                     "received from client".format(wid, request))

        # --- serialize request making sure we always get the same key for the
        #     same request (NB: the request is a dictionary which means
        #     ordering is not guaranteed).
        req_key = "".join(["({0!s},{1!s})".format(k, request.get(k, None))
                           for k in SORTED_KEYS])
        req_key = hashlib.sha1(req_key.encode("utf-8")).hexdigest()
        req_args = [request.get(key, None) for key in SORTED_KEYS
                    if request.get(key, None) is not None]
        logger_fmt = "{0:3d} {{0:>8s}} {1!r}:".format(wid, req_args)

        while True:
            resp = cache.get(req_key)
            if not isinstance(resp, CacheLocked):
                break
            time.sleep(0.25)

        if resp is None:
            logger.info(logger_fmt.format("missing"))
            logger.debug("{0:d}  key "
                         "'{1:s}' not found in memcache".format(wid, req_key))

            cache.set(req_key, CacheLocked(), LOCK_TIMEOUT)
            req_type = request.pop("type")
            real_time = request.pop("RT", False)

            # --- to avoid memory leaks it is important to call
            #     traceback.clear_frames whenever we break the exception chain
            #     by returning exceptions. traceback.clear_frames will remove
            #     local variables and will clear reference cycles.
            try:
                if req_type == "BDP":
                    resp = bbg_com_api.BbgBDP(**request)
                elif req_type == "BDH":
                    resp = bbg_com_api.BbgBDH(**request)
                else:
                    raise ValueError("Unrecognized "
                                     "request type {0:s}".format(req_type))
            except BloombergFatal as err:
                logger.critical("{0:d}  skipping "
                                "request on BloombergFatal".format(wid))
                err.exc_info = traceback.format_exc()
                socket.send_pyobj(err, flags=zmq.NOBLOCK)
                traceback.clear_frames(err.__traceback__)
                continue
            except Exception as err:
                err.exc_info = traceback.format_exc()
                resp = err
                traceback.clear_frames(err.__traceback__)
            else:
                logger.debug("{0:d}  query to "
                             "bloomberg completed successfully".format(wid))

            if real_time:
                expiry = RT_TIMEOUT
            else:
                expiry = Date.today().eod().timestamp()

            cache.set(req_key, resp, expiry)

        else:
            logger.info(logger_fmt.format("cached"))

        try:
            socket.send_pyobj(resp, flags=zmq.NOBLOCK)
            logger.debug("{0:d}  response sent to client".format(wid))
        except:
            logger.critical("{0:d}  failed to send "
                            "response to client".format(wid), exc_info=True)


###############################################################################
class BbgDataServer(object):
    # -------------------------------------------------------------------------
    def __init__(self, address=None, port=None, mc_servers=None):

        config = load_system_configuration()

        address = address or config.get("datafeed", "queue_address")
        port = port or config.getint("datafeed", "dealer_port")

        if mc_servers is None:
            mc_servers = [config.get("memcache", "url")]

        self.mc_servers = mc_servers
        self.conn_str = "tcp://{0:s}:{1:d}".format(address, port)

    # -------------------------------------------------------------------------
    def master_logger(self):
        logger = logging.getLogger(__name__)
        logger.info("starting master logger")

        socket = zmq.Context().socket(zmq.PULL)
        socket.bind("tcp://127.0.0.1:10001")

        while True:
            loginfo = socket.recv_pyobj()
            try:
                log_method = getattr(logger, loginfo.level)
            except AttributeError:
                # FIXME: do something better here
                raise
            else:
                try:
                    log_method(loginfo.msg,
                               exc_info=loginfo.exc_info, extra=loginfo.extra)
                except:
                    # FIXME: do something better here too
                    print(traceback.print_exc())
                    pass

    # -------------------------------------------------------------------------
    def start(self):
        logger = logging.getLogger(__name__)
        logger.info("BbgDataServer started, "
                    "listening on {0:s}".format(self.conn_str))

        # --- start master-logger in a dedicated thread
        thread = threading.Thread(target=self.master_logger)
        thread.daemon = True
        thread.start()

        while True:
            # --- launch pool of worker processes
            procs = []
            for k in range(NPROCS):
                args = (k, self.conn_str, self.mc_servers)
                proc = multiprocessing.Process(target=worker, args=args)
                proc.daemon = True
                proc.start()
                procs.append(thread)

            for proc in procs:
                proc.join()

            logger.info("BbgDataServer: all workers have been stopped...")

        logger.info("BbgDataServer stopped")
